# TODO a lot of test improvement is needed. Here are a few suggestions
# TODO use synthetic generated data instead of some real data in a lot of cases
# TODO benchmark vs other codes apart from CASM (ie icet, ATAT, etc)
# TODO check all cases for each function, including exception raising.
